// Sample code for pseudo-inverting a singular or otherwise bad matrix
// A usage example for the LinAlg package
//
// Suppose we need to find a pseudo-inverse of a non-square singular matrix
// A=[0 0 1 4
//    0 0 1 3
//    0 0 0 0];
//
// The example is due to John Anderson
//
// The pseudo-inverse of A is found by "inverting" the SVD of A.
// To be more precise, we use SVD to solve the equation 
// AX=B where B is a unit matrix.
//
// LinALg SVD accepts only matrices with the number of rows greater
// than the number of columns. Matrix A does not satisfy this stipulation,
// and will be rejected by the svd() constructor. We note however
// that for any matrix, transpose(inverse(A)) is inverse(transpose(A)).
// transpose(A) will be accepted by the SVD code.
//
// After we compute the pseudo-inverse, we verify the Moore-Penrose
// conditions: Matrix X is a pseudo-inverse of A iff
//	AXA = A
//	XAX = X
//	XA = (XA)' (i.e., XA is symmetric)
//	AX = (AX)' (i.e., AX is symmetric)
//
// $Id: tpseudo-inv.cc,v 1.2 2002/11/21 03:06:23 oleg Exp oleg $

#include "LAStreams.h"
#include "svd.h"
#if defined(__GNUC__) || __GNUC__ >= 3
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#else
#include <iostream.h>
#endif


int main(void)
{
 cout << "\n\nAn application of SVD to pseudo-inverting a non-square singular matrix" << endl;

 // First we allocate and fill in the Matrix A
 enum {nrows = 3, ncols = 4};
 Matrix A(nrows,ncols);

 // Data for the matrix. Note, the values are arranged column-wise!
 const float A_data [nrows*ncols] =
 {0,  0,  0,
  0,  0,  0,
  1,  1,  0,
  4,  3,  0};

 // Fill in the matrix using a Matrix stream. The stream traverses
 // the matrix in a _column_ major order.
 // In a real program, we will get data from a file rather than
 // from an array A_data
 {
   const float * Ap = A_data;
   LAStreamOut str(A);
   while( !str.eof() )
     str.get() = *Ap++;
   assert( Ap == A_data + nrows*ncols ); // Verify that we have read
					 // all data values, and no more
 }

 // Print out the matrix
 cout << "Finding a pseudo-inverse for" << endl;
 A.print("Matrix A");

 // The SVD solution itself. We do SVD on the transposed matrix,
 // and then transpose the found solution again.
 SVD svd(transposed(A));	// Perform the SVD decomposition
		// Print out the condition number of A, for information only
 cout << "\ncondition number " << svd.q_cond_number() << endl;
		// Compute the inverse as a solution of
		// A*Ainv = B
 Matrix Ainv = transposed(SVD_inv_mult(svd,Matrix(ncols,ncols).unit_matrix()));
 Ainv.print("The psedo-inverse");	// Print the solution

 cout << "\nChecking the Moore-Penrose conditions for the Ainv" << endl;
 cout << "\nVerify that Ainv * A * Ainv is Ainv" << endl;
 verify_matrix_identity(Ainv * A * Ainv, Ainv);
 cout << "\nVerify that A * Ainv * A is A" << endl;
 verify_matrix_identity(A * Ainv * A, A);
 cout << "\nVerify that Ainv * A is symmetric" << endl;
 verify_matrix_identity(Ainv * A, transposed(Ainv*A));
 cout << "\nVerify that A * Ainv is symmetric" << endl;
 verify_matrix_identity(A * Ainv, transposed(A*Ainv));

 cout << "\nAll done" << endl;
 return 0;
}


