// Sample code for solving Ax=b
// A usage example for the LinAlg package
//
// Suppose we need to solve Ax=b where
// A=[1.1  0.0 2.1         and b=[1.9
//   2.0  2.1 3.0                5.3
//   0.1  1.0 5.0                3.2
//   4.1  3.0 1.0]               2.1]
//
// The example is due to Elias Kyriakides.
//
// $Id: tslesing.cc,v 1.2 2002/11/21 03:05:45 oleg Exp oleg $

#include "LAStreams.h"
#include "svd.h"
#if defined(__GNUC__) || __GNUC__ >= 3
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#else
#include <iostream.h>
#endif


int main(void)
{
 cout << "\n\nAn application of SVD to solving sets of simultaneous "
         "equations" << endl;

 // First we allocate and fill in the Matrix A
 enum {nrows = 4, ncols = 3};
 Matrix A(nrows,ncols);

 // Data for the matrix. Note, the values are arranged column-wise!
 const float A_data [nrows*ncols] =
 {1.1, 2.0, 0.1, 4.1,
  0.0, 2.1, 1.0, 3.0,
  2.1, 3.0, 5.0, 1.0};

 // Fill in the matrix using a Matrix stream. The stream traverses
 // the matrix in a _column_ major order.
 // In a real program, we will get data from a file rather than
 // from an array A_data
 {
   const float * Ap = A_data;
   LAStreamOut str(A);
   while( !str.eof() )
     str.get() = *Ap++;
   assert( Ap == A_data + nrows*ncols ); // Verify that we have read
					 // all data values, and no more
 }

 // Now allocate and fill in the vector b.
 Vector b(1,4,	// the lower and the upper bounds for the Vector's index
	  1.9,	// Vector's data
	  5.3,
	  3.2,
	  2.1,
	  "END");


 // Print out the matrix and the vector
 cout << "Solving Ax=b for";
 A.print("Matrix A");
 b.print("vector b");

 // The SVD solution itself
 SVD svd(A);	// Perform the SVD decomposition
		// Print out the condition number of A, for information only
 cout << "\ncondition number " << svd.q_cond_number() << endl;
		// Compute the solutions vector x
 Vector x = SVD_inv_mult(svd,b);
 x.print("Solution");	// Print the solution vector

 // Print the solution vector x again, in a different way, to illustrate
 // how we can access its elements one-by-one.
 {
   cout << "The solution vector, again [" << endl;
   LAStreamIn istr(x);
   while (!istr.eof())
     cout << istr.get() << "\n";
   cout << "]" << endl;
   // Note, you can also access vector x elements using x(1)...x(4)
   // notation. Using Streams however is more efficient.
 }

 // Verify the found solution: Compare A*x with b
 // Note A is not a square matrix. Therefore, Ax is close to b, but
 // not identical to.
 Vector ax = A*x;
 ax.print("Vector A*x");
 
 // Compute the solution in a different way
 cout << "\nComputing the solution in a different way:" << endl;
 cout << "Ax=b ==> x = inverse(A'*A)*(A'*b)" << endl;
 Matrix At = transposed(A);
 Vector x_another = ((Matrix)inverse(At*A)) * (At*b);
 x_another.print("Solution inverse(A'*A)*(A'*b)");

 cout << "\nAll done" << endl;
 return 0;
}
